/*
 * Daemon which 'calls home' periodically, so that endpoints on dynamic
 * addresses (e.g. assigned via DHCP) can be found.
 * This is a simple mechanism for use on private networks, e.g. where
 * dynamic DNS would be too awkward to implement.
 * The packets are sent via syslog to a host of your choice. Since this
 * is a non-privileged daemon the source packet will not be 514, so your
 * syslog daemon must be configured to accept packets from any source port.
 */

/*
TODO:
- use Ericsson's snmp 'heartbeat' trap instead of syslog?
- different first message / subsequent message?
*/

#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/ether.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <time.h>
#include <unistd.h>

/*
 * Get MAC address of interface into static buffer
 */

static char *get_macaddr(const char *ifname)
{
  char *str = "<unknown>";
  struct ifreq req;
  int s = socket(PF_INET, SOCK_DGRAM, IPPROTO_IP);

  if (s < 0) return str;
  memset(&req, 0, sizeof(req));
  strncpy(req.ifr_name, ifname, sizeof(req.ifr_name));
  if (ioctl(s, SIOCGIFHWADDR, &req) == 0)
    str = ether_ntoa(&req.ifr_hwaddr.sa_data);
  else
    perror("SIOCGIFHWADDR");
  close(s);
  return str;
}

/*
 * Expand string, into a static buffer since I'm lazy
 *
 * %%       : %
 * %t       : timestamp (syslog format)
 * %m{eth0} : MAC address of eth0
 * %f{fname}: insert first line of file, minus trailing \n
 * %l       : LOG_INFO | LOG_USER
 */

static char *addstr(char *bp, char *const be, const char *str)
{
  while (*str && bp < be)
    *bp++ = *str++;
  return bp;
}

static char *expand_string(char *str)
{
  static char buf[1024];
  char *bp = buf;
  char *const be = buf + sizeof(buf) - 1;
  
  while(bp < be) {
    int ch = *str++;
    if (ch == '%') {
      char arg[128];
      char *p;

      ch = *str++;
      if (!ch) break;

      if (*str == '{' && (p = strchr(str, '}')) != 0 && p-str <= sizeof(arg)) {
        strncpy(arg, str+1, p-str-1);
        arg[p-str-1] = '\0';
        str = p+1;
      }
      else
        arg[0] = '\0';

      if (ch == '%') {
        *bp++ = ch;
      }
      else if (ch == 'l') {
        snprintf(bp, be-bp, "%d", LOG_INFO | LOG_USER);
        while(*bp && bp<be) bp++;
      }
      else if (ch == 't') {
        time_t now = time(0);
        char *timbuf = ctime(&now);
        timbuf[19] = '\0';
        bp = addstr(bp, be, timbuf+4);
      }
      else if (ch == 'm' && arg[0]) {
        char *mac = get_macaddr(arg);
        bp = addstr(bp, be, mac);
      }
      else if (ch == 'f' && arg[0]) {
        FILE *f = fopen(arg, "r");
        if (f) {
          while ((ch = fgetc(f)) >= ' ' && bp < be)
            *bp++ = ch;
          fclose(f);
        }
        continue;
      }
      else if (!ch)
        break;
      /* else ignore */
    }
    else if (!ch)
      break;
    else
      *bp++ = ch;
  }
  *bp = '\0';
  return buf;
}

static void usage(void)
{
  fprintf(stderr, "Usage: beacon [-i interval] [-I initial-delay] [-h host] [-s svc] [-m msg]\n");
  exit(1);
}

static int opensock(const char *hostname, const char *service)
{
  struct addrinfo hints, *res, *res0;
  int err;
  int s;
  const char *errstr = "";

  memset(&hints, 0, sizeof(hints));
  hints.ai_family = PF_UNSPEC;
  hints.ai_socktype = SOCK_DGRAM;
  if ((err = getaddrinfo(hostname, service, &hints, &res0)) != 0) {
    fprintf(stderr, "%s: %s\n", hostname, gai_strerror(err));
    return -1;
  }
  s = -1;
  for (res = res0; res; res = res->ai_next) {
    s = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
    if (s < 0) {
      errstr = "socket";
      continue;
    }
    if (connect(s, res->ai_addr, res->ai_addrlen) < 0) {
      close(s);
      s = -1;
      errstr = "connect";
      continue;
    }
    break;
  }
  if (s < 0) {
    perror(errstr);
    return -1;
  }
  freeaddrinfo(res0);
  return s;
}

int main(int argc, char *argv[])
{
  int delay = 0;	/* until first message sent */
  int interval = 3600;	/* until subsequent messages sent */
  char *host = "localhost";
  char *svc = "syslog";
  char *msg = "<%l>%t [%m{eth0}] beacon uptime %f{/proc/uptime}";
  int ch;
  int s = -1;

  while ((ch = getopt(argc, argv, "i:I:h:s:m:")) >= 0) {
    switch(ch) {
    case 'I':
      delay = atoi(optarg);
      if (delay < 0) usage();
      break;
    case 'i':
      interval = atoi(optarg);
      if (interval <= 0) usage();
      break;
    case 'h':
      host = strdup(optarg);
      break;
    case 's':
      svc = strdup(optarg);
      break;
    case 'm':
      msg = strdup(optarg);
      break;
    default:
      usage();
    }
  }
  argc -= optind;
  argv += optind;
  if (argc != 0) usage();

  while(1) {
    sleep(delay);
    
    if (s < 0) s = opensock(host, svc);
    if (s >= 0) {
      char *str = expand_string(msg);
      if (send(s, str, strlen(str), 0) < 0) {
        close(s);
        s = -1;
      };
    }
    delay = interval;
  }
}
